import { DefaultRequestOptions, FullRequestOptions, FullResponse, RequestOptions } from '@orbit/data';
import IdentityMap from '@orbit/identity-map';
import { MemoryCache } from '@orbit/memory';
import { RecordCacheQueryOptions, RecordCacheTransformOptions } from '@orbit/record-cache';
import { InitializedRecord, RecordIdentity, RecordKeyMap, RecordOperation, RecordQueryResult, RecordSchema, RecordTransformResult, StandardRecordValidator, UninitializedRecord } from '@orbit/records';
import { StandardValidator, ValidatorForFn } from '@orbit/validators';
import type Store from './store';
import LiveQuery from './live-query';
import Model from './model';
import { ModelAwareQueryBuilder, ModelAwareQueryOrExpressions, ModelAwareTransformBuilder, ModelAwareTransformOrOperations, RecordIdentityOrModel } from './utils/model-aware-types';
import { ModelFields } from './utils/model-fields';
export interface CacheSettings {
    sourceCache: MemoryCache;
    store?: Store;
}
export default class Cache {
    #private;
    allowUpdates: boolean;
    protected _identityMap: IdentityMap<RecordIdentity, Model>;
    constructor(settings: CacheSettings);
    get sourceCache(): MemoryCache;
    get store(): Store | undefined;
    get keyMap(): RecordKeyMap | undefined;
    get schema(): RecordSchema;
    get queryBuilder(): ModelAwareQueryBuilder;
    get transformBuilder(): ModelAwareTransformBuilder;
    get validatorFor(): ValidatorForFn<StandardValidator | StandardRecordValidator> | undefined;
    get defaultQueryOptions(): DefaultRequestOptions<RecordCacheQueryOptions> | undefined;
    set defaultQueryOptions(options: DefaultRequestOptions<RecordCacheQueryOptions> | undefined);
    get defaultTransformOptions(): DefaultRequestOptions<RecordCacheTransformOptions> | undefined;
    set defaultTransformOptions(options: DefaultRequestOptions<RecordCacheTransformOptions> | undefined);
    getRecordData(type: string, id: string): InitializedRecord | undefined;
    includesRecord(type: string, id: string): boolean;
    recordIdFromKey(type: string, keyName: string, keyValue: string): string;
    /**
     * @deprecated
     */
    peekRecordData(type: string, id: string): InitializedRecord | undefined;
    /**
     * @deprecated
     */
    peekRecord(type: string, id: string): Model | undefined;
    /**
     * @deprecated
     */
    peekRecords(type: string): Model[];
    /**
     * @deprecated
     */
    peekRecordByKey(type: string, key: string, value: string): Model | undefined;
    /**
     * @deprecated
     */
    peekKey(identity: RecordIdentity, key: string): string | undefined;
    /**
     * @deprecated
     */
    peekAttribute(identity: RecordIdentity, attribute: string): any;
    /**
     * @deprecated
     */
    peekRelatedRecord(identity: RecordIdentity, relationship: string): Model | null | undefined;
    /**
     * @deprecated
     */
    peekRelatedRecords(identity: RecordIdentity, relationship: string): Model[] | undefined;
    update<RequestData extends RecordTransformResult<Model> = RecordTransformResult<Model>>(transformOrOperations: ModelAwareTransformOrOperations, options?: DefaultRequestOptions<RequestOptions>, id?: string): RequestData;
    update<RequestData extends RecordTransformResult<Model> = RecordTransformResult<Model>>(transformOrOperations: ModelAwareTransformOrOperations, options: FullRequestOptions<RequestOptions>, id?: string): FullResponse<RequestData, unknown, RecordOperation>;
    query<RequestData extends RecordQueryResult<Model> = RecordQueryResult<Model>>(queryOrExpressions: ModelAwareQueryOrExpressions, options?: DefaultRequestOptions<RecordCacheQueryOptions>, id?: string): RequestData;
    query<RequestData extends RecordQueryResult<Model> = RecordQueryResult<Model>>(queryOrExpressions: ModelAwareQueryOrExpressions, options: FullRequestOptions<RecordCacheQueryOptions>, id?: string): FullResponse<RequestData, undefined, RecordOperation>;
    liveQuery(queryOrExpressions: ModelAwareQueryOrExpressions, options?: DefaultRequestOptions<RecordCacheQueryOptions>, id?: string): LiveQuery;
    /**
     * @deprecated
     */
    find(type: string, id?: string): Model | Model[] | undefined;
    findRecord(type: string, id: string, options?: DefaultRequestOptions<RecordCacheQueryOptions>): Model | undefined;
    findRecord(identity: RecordIdentityOrModel, options?: DefaultRequestOptions<RecordCacheQueryOptions>): Model | undefined;
    findRecords(typeOrIdentities: string | RecordIdentityOrModel[], options?: DefaultRequestOptions<RecordCacheQueryOptions>): Model[];
    /**
     * Adds a record
     */
    addRecord<RequestData extends RecordTransformResult<Model> = Model>(properties: UninitializedRecord | ModelFields, options?: DefaultRequestOptions<RecordCacheQueryOptions>): RequestData;
    /**
     * Updates a record
     */
    updateRecord<RequestData extends RecordTransformResult<Model> = Model>(properties: InitializedRecord | ModelFields, options?: DefaultRequestOptions<RecordCacheQueryOptions>): RequestData;
    /**
     * Removes a record
     */
    removeRecord(identity: RecordIdentityOrModel, options?: DefaultRequestOptions<RecordCacheQueryOptions>): void;
    unload(identity: RecordIdentity): void;
    lookup(identity: RecordIdentity): Model;
    _lookupQueryResult(result: RecordQueryResult<InitializedRecord>, isArray: boolean): RecordQueryResult<Model>;
    _lookupTransformResult(result: RecordTransformResult<InitializedRecord>, isArray: boolean): RecordTransformResult<Model>;
    private lookupQueryExpressionResult;
    private lookupOperationResult;
    private notifyPropertyChange;
    private generatePatchListener;
}
